# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Step statistics."""

import dataclasses
from typing import Optional

# Note:`*_end_time` fields are not required here as it can be calculated by
# *_start_time + *_duration_secs


@dataclasses.dataclass
class SaveStepStatistics:
  """Attributes for save step statistics.

  Attributes:
    step: The step number.
    event_type: The event type.
    checkpoint_manager_blocking_start_time: The start time of checkpoint manager
      blocking section.
    directory: The directory of the checkpoint.
    reached_preemption: Whether the event reached preemption.
    preemption_received_at: The time when preemption was received.
    wait_for_prev_start_time: The start time of waiting for previous checkpoint.
    checkpointer_blocking_start_time: The start time of blocking time introduced
      by checkpointer.
    get_old_steps_start_time: The start time of getting old steps.
    synchronous: Whether the event is synchronous.
    wait_for_prev_duration_secs: The duration of waiting for previous
      checkpoint.
    checkpointer_blocking_duration_secs: The duration of blocking time
      introduced by checkpointer.
    get_old_steps_duration_secs: The duration of getting old steps.
    checkpoint_manager_blocking_duration_secs: The duration of checkpoint
      manager blocking section.
  """

  step: Optional[int] = None
  event_type: Optional[str] = "save"
  directory: Optional[str] = None
  reached_preemption: Optional[bool] = False
  preemption_received_at: Optional[float] = None
  synchronous: Optional[bool] = False
  wait_for_prev_start_time: Optional[float] = None
  wait_for_prev_duration_secs: Optional[float] = None
  checkpointer_blocking_start_time: Optional[float] = None
  checkpointer_blocking_duration_secs: Optional[float] = None
  get_old_steps_start_time: Optional[float] = None
  get_old_steps_duration_secs: Optional[float] = None
  checkpoint_manager_blocking_start_time: Optional[float] = None
  checkpoint_manager_blocking_duration_secs: Optional[float] = None


@dataclasses.dataclass
class RestoreStepStatistics:
  """Attributes for restore step statistics.

  Attributes:
    step: The step number.
    event_type: The event type.
    directory: The directory of the checkpoint.
    checkpointer_start_time: The start time of restoring the checkpoint, while
      using the checkpointer.
    checkpointer_duration_secs: The total duration for restoring the checkpoint,
      while using the checkpointer.
    checkpoint_manager_start_time: The start time for restoring the checkpoint,
      while using the checkpoint manager.
    checkpoint_manager_duration_secs: The total duration for restoring the
      checkpoint, while using the checkpoint manager.
  """

  step: Optional[int] = None
  event_type: Optional[str] = "restore"
  directory: Optional[str] = None
  event_type: Optional[str] = "restore"
  directory: Optional[str] = None
  checkpointer_start_time: Optional[float] = None
  checkpointer_duration_secs: Optional[float] = None
  checkpoint_manager_start_time: Optional[float] = None
  checkpoint_manager_duration_secs: Optional[float] = None


@dataclasses.dataclass
class EmergencyRestoreStepStatistics:
  """Attributes for emergency restore step statistics.

  Attributes:
    step: The step number.
    event_type: The event type.
    checkpoint_manager_start_time: The start time of checkpoint manager
      restore event.
    directory: The directory of the checkpoint.
    is_restoring_slice: Whether the event takes place on the slice responsible
      for reading from the storage location. (Note that in_primary_slice=True
      necessarily implies is_restoring_slice=True.)
    in_primary_slice: Whether the event takes place on the slice designated as
      primary (responsible for restoring from persistent storage).
    checkpointer_start_time: The start time of restoring the checkpoint, while
      using the checkpointer.
    checkpointer_duration_secs: The total duration for restoring the checkpoint,
      while using the checkpointer.
    broadcast_start_time: The start time of broadcasting(Restore).The broadcast
      operation performed by SingleReplicaArrayHandler won't be captured in this
      context.
    broadcast_duration_secs: The duration of broadcasting(Restore).
    checkpoint_manager_duration_secs: The total duration of checkpoint
      manager restore event.
  """

  step: Optional[int] = None
  event_type: Optional[str] = "emergency_restore"
  checkpoint_manager_start_time: Optional[float] = None
  directory: Optional[str] = None
  is_restoring_slice: Optional[bool] = False
  in_primary_slice: Optional[bool] = False
  checkpointer_start_time: Optional[float] = None
  checkpointer_duration_secs: Optional[float] = None
  broadcast_start_time: Optional[float] = None
  broadcast_duration_secs: Optional[float] = None
  checkpoint_manager_duration_secs: Optional[float] = None



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main module for PyTree checkpoint metadata storage."""

from __future__ import annotations

import asyncio
import collections
import dataclasses
import enum
import functools
import operator
from typing import Any, Dict, Hashable, List, Optional, Tuple, TypeAlias, TypeVar, Union

from absl import logging
from etils import epath
import jax
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import pytree_metadata_options as pytree_metadata_options_lib
from orbax.checkpoint._src.metadata import tree_rich_types
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.metadata import value_metadata_entry
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
from orbax.checkpoint._src.serialization import types
from orbax.checkpoint._src.tree import utils as tree_utils

ValueMetadataEntry: TypeAlias = value_metadata_entry.ValueMetadataEntry
PyTreeMetadataOptions: TypeAlias = (
    pytree_metadata_options_lib.PyTreeMetadataOptions
)
PyTree: TypeAlias = Any
KeyEntry = TypeVar('KeyEntry', bound=Hashable)
KeyPath: TypeAlias = tuple[KeyEntry, ...]

PYTREE_METADATA_OPTIONS = pytree_metadata_options_lib.PYTREE_METADATA_OPTIONS

_KEY_NAME = 'key'
_KEY_TYPE = 'key_type'
_TREE_METADATA_KEY = 'tree_metadata'
_KEY_METADATA_KEY = 'key_metadata'
_VALUE_METADATA_KEY = 'value_metadata'
_USE_ZARR3 = 'use_zarr3'
_STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE = 'store_array_data_equal_to_fill_value'
_VALUE_METADATA_TREE = 'value_metadata_tree'


class KeyType(enum.Enum):
  """Enum representing PyTree key type."""

  SEQUENCE = 1
  DICT = 2

  def to_json(self) -> int:
    return self.value

  @classmethod
  def from_json(cls, value: int) -> KeyType:
    return cls(value)


def _get_key_metadata_type(key: Any) -> KeyType:
  """Translates the JAX key class into a proto enum."""
  if tree_utils.is_sequence_key(key):
    return KeyType.SEQUENCE
  elif tree_utils.is_dict_key(key):
    return KeyType.DICT
  else:
    raise ValueError(f'Unsupported KeyEntry: {type(key)}: "{key}"')


def _keypath_from_key_type(key_name: str, key_type: KeyType) -> Any:
  """Converts from Key in InternalTreeMetadata to JAX keypath class."""
  if key_type == KeyType.SEQUENCE:
    return jax.tree_util.SequenceKey(int(key_name))
  elif key_type == KeyType.DICT:
    return jax.tree_util.DictKey(key_name)
  else:
    raise ValueError(f'Unsupported KeyEntry: {key_type}')


@dataclasses.dataclass
class NestedKeyMetadataEntry:
  """Represents a key at a single level of nesting."""

  nested_key_name: str
  key_type: KeyType

  def to_json(self) -> Dict[str, Union[str, int]]:
    return {
        _KEY_NAME: self.nested_key_name,
        _KEY_TYPE: self.key_type.to_json(),
    }

  @classmethod
  def from_json(
      cls, json_dict: Dict[str, Union[str, int]]
  ) -> NestedKeyMetadataEntry:
    return NestedKeyMetadataEntry(
        nested_key_name=json_dict[_KEY_NAME],
        key_type=KeyType.from_json(json_dict[_KEY_TYPE]),
    )


@dataclasses.dataclass
class KeyMetadataEntry:
  """Represents metadata for a key (all levels of nesting)."""

  nested_key_metadata_entries: List[NestedKeyMetadataEntry]

  def to_json(self) -> Tuple[Dict[str, Union[str, int]], ...]:
    return tuple(
        [entry.to_json() for entry in self.nested_key_metadata_entries]
    )

  @classmethod
  def from_json(
      cls, json_dict: Tuple[Dict[str, Union[str, int]], ...]
  ) -> KeyMetadataEntry:
    return KeyMetadataEntry(
        [NestedKeyMetadataEntry.from_json(entry) for entry in json_dict]
    )

  @classmethod
  def build(cls, keypath: KeyPath) -> KeyMetadataEntry:
    return KeyMetadataEntry([
        NestedKeyMetadataEntry(
            str(tree_utils.get_key_name(k)), _get_key_metadata_type(k)
        )
        for k in keypath
    ])


@dataclasses.dataclass
class InternalTreeMetadataEntry:
  """Represents metadata for a named key/value pair in a tree."""

  keypath: str
  key_metadata: KeyMetadataEntry
  value_metadata: ValueMetadataEntry

  def to_json(self) -> Dict[str, Any]:
    return {
        self.keypath: {
            _KEY_METADATA_KEY: self.key_metadata.to_json(),
            _VALUE_METADATA_KEY: self.value_metadata.to_json(),
        }
    }

  @classmethod
  def from_json(
      cls, keypath: str, json_dict: Dict[str, Any]
  ) -> InternalTreeMetadataEntry:
    return InternalTreeMetadataEntry(
        keypath,
        KeyMetadataEntry.from_json(json_dict[_KEY_METADATA_KEY]),
        ValueMetadataEntry.from_json(json_dict[_VALUE_METADATA_KEY]),
    )

  @classmethod
  def build(
      cls,
      keypath: KeyPath,
      info: types.ParamInfo,
      save_arg: types.SaveArgs,
  ) -> InternalTreeMetadataEntry:
    """Builds a InternalTreeMetadataEntry."""
    return InternalTreeMetadataEntry(
        keypath=str(tuple([str(tree_utils.get_key_name(k)) for k in keypath])),
        key_metadata=KeyMetadataEntry.build(keypath),
        value_metadata=ValueMetadataEntry.build(info, save_arg),
    )

  def jax_keypath(self) -> KeyPath:
    keypath = []
    for nested_key_entry in self.key_metadata.nested_key_metadata_entries:
      nested_key_name = nested_key_entry.nested_key_name
      key_type = nested_key_entry.key_type
      keypath.append(_keypath_from_key_type(nested_key_name, key_type))
    return tuple(keypath)


@dataclasses.dataclass
class InternalTreeMetadata:
  """Metadata representation of a PyTree."""

  tree_metadata_entries: List[InternalTreeMetadataEntry]
  use_zarr3: bool
  store_array_data_equal_to_fill_value: bool
  pytree_metadata_options: PyTreeMetadataOptions
  value_metadata_tree: PyTree | None = None

  def __post_init__(self):
    logging.info(
        'Created InternalTreeMetadata with pytree_metadata_options= %s',
        self.pytree_metadata_options,
    )

  @classmethod
  def build(
      cls,
      param_infos: PyTree,
      *,
      save_args: Optional[PyTree] = None,
      use_zarr3: bool = False,
      pytree_metadata_options: PyTreeMetadataOptions = (
          PYTREE_METADATA_OPTIONS
      ),
  ) -> InternalTreeMetadata:
    """Returns an InternalTreeMetadata instance."""
    if save_args is None:
      save_args = jax.tree.map(
          lambda _: types.SaveArgs(),
          param_infos,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
    flat_info_with_keys, _ = jax.tree_util.tree_flatten_with_path(
        param_infos, is_leaf=tree_utils.is_empty_or_leaf
    )
    flat_save_args_with_keys, _ = jax.tree_util.tree_flatten_with_path(
        save_args, is_leaf=tree_utils.is_empty_or_leaf
    )
    tree_metadata_entries = []
    for (keypath, info), (_, save_arg) in zip(
        flat_info_with_keys, flat_save_args_with_keys
    ):
      tree_metadata_entries.append(
          InternalTreeMetadataEntry.build(keypath, info, save_arg)
      )
    value_metadata_tree = None
    if pytree_metadata_options.support_rich_types:
      value_metadata_tree = jax.tree_util.tree_map(
          ValueMetadataEntry.build,
          param_infos,
          save_args,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
    return InternalTreeMetadata(
        tree_metadata_entries,
        use_zarr3,
        ts_utils.STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE,
        pytree_metadata_options,
        value_metadata_tree,
    )

  def to_json(self) -> Dict[str, Any]:
    """Returns a JSON representation of the metadata.

    Uses JSON format::
    ```
      {
          _TREE_METADATA_KEY: {
            "(top_level_key, lower_level_key)": {
                _KEY_METADATA_KEY: (
                    {_KEY_NAME: "top_level_key", _KEY_TYPE: <KeyType
                    (int)>},
                    {_KEY_NAME: "lower_level_key", _KEY_TYPE: <KeyType
                    (int)>},
                )
                _VALUE_METADATA_KEY: {
                    _VALUE_TYPE: "jax.Array",
                    _SKIP_DESERIALIZE: True/False,
                }
            }
            ...
          },
          _USE_ZARR3: True/False,
          _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE: True,
          _VALUE_METADATA_TREE: '{
            "mu_nu": {
              "category": "namedtuple",
              "module": "orbax.checkpoint._src.testing.test_tree_utils",
              "clazz": "MuNu",
              "entries": [
                {
                  "key": "mu",
                  "value": {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "jax.Array",
                      "skip_deserialize": false
                    }
                  }
                },
                {
                  "key": "nu",
                  "value": {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "np.ndarray",
                      "skip_deserialize": false
                    }
                  }
                }
              ]
            },
            "my_tuple": {
              "category": "custom",
              "clazz": "tuple",
              "entries": [
                  {
                    "category": "custom",
                    "clazz": "ValueMetadataEntry",
                    "data": {
                      "value_type": "np.ndarray",
                      "skip_deserialize": false
                    }
                  }
              ]
            }
          }'
      }
    ```
    """
    json_object = {
        _TREE_METADATA_KEY: functools.reduce(
            operator.ior,
            [entry.to_json() for entry in self.tree_metadata_entries],
            {},
        ),
        _USE_ZARR3: self.use_zarr3,
        _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE: (
            self.store_array_data_equal_to_fill_value
        ),
    }
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    if (
        self.pytree_metadata_options.support_rich_types
        and self.value_metadata_tree is not None
    ):
      json_object[_VALUE_METADATA_TREE] = (
          tree_rich_types.value_metadata_tree_to_json_str(
              self.value_metadata_tree
          )
      )
    return json_object

  @classmethod
  def from_json(
      cls,
      json_dict: Dict[str, Any],
      pytree_metadata_options: PyTreeMetadataOptions = (
          PYTREE_METADATA_OPTIONS
      ),
  ) -> InternalTreeMetadata:
    """Returns an InternalTreeMetadata instance from its JSON representation."""
    use_zarr3 = json_dict.get(_USE_ZARR3, False)
    store_array_data_equal_to_fill_value = json_dict.get(
        _STORE_ARRAY_DATA_EQUAL_TO_FILL_VALUE, False
    )

    tree_metadata_entries = []
    for keypath, json_tree_metadata_entry in json_dict[
        _TREE_METADATA_KEY
    ].items():
      tree_metadata_entries.append(
          InternalTreeMetadataEntry.from_json(keypath, json_tree_metadata_entry)
      )
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    value_metadata_tree = None
    if (
        pytree_metadata_options.support_rich_types
        and _VALUE_METADATA_TREE in json_dict
    ):
      value_metadata_tree = tree_rich_types.value_metadata_tree_from_json_str(
          json_dict[_VALUE_METADATA_TREE]
      )
    return InternalTreeMetadata(
        tree_metadata_entries,
        use_zarr3=use_zarr3,
        pytree_metadata_options=pytree_metadata_options,
        value_metadata_tree=value_metadata_tree,
        store_array_data_equal_to_fill_value=store_array_data_equal_to_fill_value,
    )

  def as_nested_tree(self) -> Dict[str, Any]:
    """Converts to a nested tree, with leaves of ValueMetadataEntry."""
    # TODO: b/365169723 - Support versioned evolution of metadata storage.
    if (
        self.pytree_metadata_options.support_rich_types
        and self.value_metadata_tree is not None
    ):
      return self.value_metadata_tree

    return tree_utils.from_flattened_with_keypath([
        (entry.jax_keypath(), entry.value_metadata)
        for entry in self.tree_metadata_entries
    ])

  def as_user_metadata(
      self,
      directory: epath.Path,
      type_handler_registry: types.TypeHandlerRegistry,
      *,
      use_ocdbt: bool = True,
  ) -> PyTree:
    """Returns a user-facing PyTree with leaves of `ValueMetadataEntry`.

    The returned PyTree conforms to the structure of `self` InternalTreeMetadata
    but the `ValueMetadataEntry` leaves are derived from the checkpoints in
    `directory`.

    Args:
      directory: The directory to read the checkpoint from.
      type_handler_registry: `TypeHandlerRegistry` whose registered
        TypeHandlers' metadata are used to build the `ValueMetadataEntry`
        leaves. See `TypeHandler.metadata()`.
      use_ocdbt: Whether to use OCDBT for reading the metadata from tensorstore.
    """
    flat_param_infos = {}
    flat_restore_types = {}
    reference_metadata_tree = self.as_nested_tree()
    ts_context = ts_utils.get_ts_context(use_ocdbt=use_ocdbt)
    for keypath, value_meta in tree_utils.to_flat_dict(
        reference_metadata_tree
    ).items():
      param_name = '.'.join(keypath)
      flat_param_infos[keypath] = types.ParamInfo(
          name=param_name,
          path=directory / param_name,
          parent_dir=directory,
          skip_deserialize=value_meta.skip_deserialize,
          is_ocdbt_checkpoint=use_ocdbt,
          use_zarr3=self.use_zarr3,
          ts_context=ts_context,
      )
      flat_restore_types[keypath] = value_meta.value_type

    flat_metadatas = {}
    batched_param_infos = collections.defaultdict(list)
    batched_keypaths = collections.defaultdict(list)
    for keypath in flat_param_infos:
      param_info = flat_param_infos[keypath]
      restore_type = flat_restore_types[keypath]
      if param_info.skip_deserialize:
        if empty_values.is_empty_typestr(restore_type):
          flat_metadatas[keypath] = empty_values.get_empty_value_from_typestr(
              restore_type
          )
        else:
          flat_metadatas[keypath] = value_metadata.Metadata(
              name=param_info.name, directory=param_info.parent_dir
          )
      else:
        batched_keypaths[restore_type].append(keypath)
        batched_param_infos[restore_type].append(param_info)

    metadata_ops = []
    for restore_type, param_infos in batched_param_infos.items():
      handler = type_handler_registry.get(restore_type)
      metadata_ops.append(handler.metadata(param_infos))

    async def _get_metadata():
      return await asyncio.gather(*metadata_ops)

    batched_metadatas = asyncio_utils.run_sync(_get_metadata())
    for keypath_batch, metadata_batch in zip(
        batched_keypaths.values(), batched_metadatas
    ):
      for keypath, value in zip(keypath_batch, metadata_batch):
        flat_metadatas[keypath] = value
    return tree_utils.from_flat_dict(
        flat_metadatas, target=reference_metadata_tree
    )


def serialize_tree(
    tree: PyTree,
    pytree_metadata_options: PyTreeMetadataOptions,
) -> PyTree:
  """Transforms a PyTree to a serializable format.

  IMPORTANT: If `pytree_metadata_options.support_rich_types` is false, the
  returned tree replaces tuple container nodes with list nodes.

  IMPORTANT: If `pytree_metadata_options.support_rich_types` is false, the
  returned tree replaces NamedTuple container nodes with dict
  nodes.

  If `pytree_metadata_options.support_rich_types` is true, then the returned
  tree is the same as the input tree retaining empty nodes as leafs.

  Args:
    tree: The tree to serialize.
    pytree_metadata_options: `PyTreeMetadataOptions` for managing PyTree
      metadata.

  Returns:
    The serialized PyTree.
  """
  if pytree_metadata_options.support_rich_types:
    return jax.tree_util.tree_map(
        lambda x: x,
        tree,
        is_leaf=tree_utils.is_empty_or_leaf,
    )

  return tree_utils.serialize_tree(tree, keep_empty_nodes=True)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class to manage checkpoints: AbstractCheckpointManager."""

import abc
from typing import Any, Mapping, Optional, Protocol, Sequence, Union

from etils import epath
from orbax.checkpoint import args as args_lib

PyTree = Any
SaveParams = Mapping[str, Any]
RestoreParams = SaveParams


class AbstractCheckpointManager(Protocol):
  """Interface to manage checkpoints.

  Allows a user to save and restore objects for which a Checkpointer
  implementation exists (e.g. PyTreeCheckpointer for PyTrees). The class
  keeps track of multiple checkpointable objects in the following structure::

    path/to/directory/    (top-level directory)
      0/    (step)
        params/    (first saveable)
          ...
        metadata/    (second saveable)
          ...
      1/    (step)
        ...
      2/    (step)
        ...
      ...
  """

  @property
  @abc.abstractmethod
  def directory(self) -> epath.Path:
    """Returns the top-level directory containing checkpoints for all items."""

  @abc.abstractmethod
  def all_steps(self, read: bool = False) -> Sequence[int]:
    """Returns all steps tracked by the manager.

    Args:
      read: If True, forces a read directly from the storage location.
        Otherwise, a cached result can be returned.

    Returns:
      A sequence of steps (integers)
    """

  @abc.abstractmethod
  def latest_step(self) -> Optional[int]:
    """Returns the latest step saved.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """

  @abc.abstractmethod
  def best_step(self) -> Optional[int]:
    """Returns the best step saved, as defined by `options.best_fn`.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """

  @abc.abstractmethod
  def reload(self):
    """Performs disk reads to ensure internal properties are up to date."""

  @abc.abstractmethod
  def reached_preemption(self, step: int) -> bool:
    """Returns True if a preemption sync point has been reached."""

  @abc.abstractmethod
  def should_save(self, step: int) -> bool:
    """Returns True if a checkpoint should be saved for the current step.

    This depends the previous step and save interval.

    Args:
      step: int

    Returns:
      True if the checkpoint should be saved.
    """

  @abc.abstractmethod
  def delete(self, step: int):
    """Deletes a step checkpoint."""

  @abc.abstractmethod
  def save(
      self,
      step: int,
      items: Optional[Union[Any, Mapping[str, Any]]] = None,
      save_kwargs: Optional[Union[SaveParams, Mapping[str, SaveParams]]] = None,
      metrics: Optional[PyTree] = None,
      force: Optional[bool] = False,
      args: Optional[args_lib.CheckpointArgs] = None,
  ) -> bool:
    """Saves the provided items.

    This method should be called by all hosts - process synchronization and
    actions that need to be performed on only one host are managed internally.

    NOTE: The `items` and `save_kwargs` arguments are deprecated, use `args`
    instead. Make sure to configure `CheckpointManager` with `item_names`.

    `args` should be a subclass of
    `orbax.checkpoint.args.CheckpointArgs`, the specific type of which is used
    to indicate what logic is used to save the object. For a typical, PyTree of
    arrays, use `StandardSave`/`StandardRestore`.

    When constructing the `CheckpointManager`, if no `item_names` were provided,
    it is assumed that we are managing a single object. If `item_names` were
    provided, it is assumed that we are managing multiple objects, and `args`
    must be `orbax.checkpoint.args.CompositeArgs`. See below for details.

    Example::

      # Single item
      mngr = ocp.CheckpointManager(directory)
      mngr.save(step, args=ocp.args.StandardSave(my_train_state))

      # Multiple items
      mngr = ocp.CheckpointManager(directory, item_names=('state', 'meta'))
      mngr.save(step, args=ocp.args.Composite(
          state=ocp.args.StandardSave(my_train_state),
          meta=ocp.args.JsonSave(my_metadata)
      ))

    Args:
      step: current step, int
      items: a savable object, or a dictionary of object name to savable object.
      save_kwargs: save kwargs for a single Checkpointer, or a dictionary of
        object name to kwargs needed by the Checkpointer implementation to save
        the object.
      metrics: a dictionary of metric name (string) to numeric value to be
        tracked along with this checkpoint. Required if `options.best_fn` is
        set. Allows users to specify a metric value to determine which
        checkpoints are best and should be kept (in conjunction with
        `options.max_to_keep`).
      force: if `True`, this method will attempt to save a checkpoint regardless
        of the result of `AbstractCheckpointManager.should_save(step)`. By
        default, `save` will only write a checkpoint to disk when the options
        permit, e.g. when `step` is in `options.save_interval_steps` or
        `options.save_on_steps`. Setting `force=True` will not overwrite
        existing checkpoints.
      args: `CheckpointArgs` which is used to save checkpointable objects with
        the appropriate logic.

    Returns:
      bool indicating whether a save operation was performed.
    Raises:
      ValueError: if `track_best` was indicated but `metrics` is not provided.
      ValueError: directory creation failed.
      ValueError: if an item is provided for which no `Checkpointer` is
      found.
      ValueError: if the checkpoint already exists.
    """

  @abc.abstractmethod
  def restore(
      self,
      step: Optional[int],
      items: Optional[Union[Any, Mapping[str, Any]]] = None,
      restore_kwargs: Optional[
          Union[RestoreParams, Mapping[str, RestoreParams]]
      ] = None,
      directory: Optional[epath.PathLike] = None,
      args: Optional[args_lib.CheckpointArgs] = None,
  ) -> Union[Any, Mapping[str, Any], args_lib.Composite]:
    """Restores from the given step and provided items.

    This method should be called by all hosts - process synchronization and
    actions that need to be performed on only one host are managed internally.

    NOTE: The `items` and `restore_kwargs` arguments are deprecated, use `args`
    instead. Make sure to configure `CheckpointManager` with `item_names`.
    See `save` docstring for additional details.

    Example::

      # Single item
      mngr = ocp.CheckpointManager(directory)
      mngr.restore(step, args=ocp.args.StandardRestore(abstract_train_state))

      # Multiple items
      mngr = ocp.CheckpointManager(directory, item_names=('state', 'meta'))
      mngr.restore(step, args=ocp.args.Composite(
          state=ocp.args.StandardRestore(abstract_train_state),
          meta=ocp.args.JsonRestore(),
      ))
      # If it is acceptable to restore without providing additional arguments,
      # and if a save has already been performed, it is ok to do the following:
      mngr.restore(step, args=ocp.args.Composite(state=None, meta=None))
      # If a save has not already been performed, there is no way for Orbax to
      # know how to restore the objects. If a save has already been performed,
      # it remembers the logic used to save the objects.

    Args:
      step: current step, int
      items: a restoreable object, or a dictionary of object name to restorable
        object.
      restore_kwargs: restore kwargs for a single Checkpointer, or a dictionary
        of object name to kwargs needed by the Checkpointer implementation to
        restore the object.
      directory: if provided, uses the given directory rather than the
        `directory` property of this class. Can be used to restore checkpoints
        from an independent location.
      args: `CheckpointArgs` which is used to restore checkpointable objects
        with the appropriate logic.

    Returns:
      If managing a single item, returns a single checkpointable object.
      If managing multiple items, returns ocp.args.Composite, where the keys
      are item names, and values are checkpointable objects.
    """

  @abc.abstractmethod
  def item_metadata(
      self, step: int
  ) -> Union[Any, Mapping[str, Any], args_lib.Composite]:
    """For all Checkpointers, returns any metadata associated with the item.

    Calls the `metadata` method for each Checkpointer and returns a
    mapping of each item name to the restored metadata. If the manager only
    manages a single item, a single metadata will be returned instead.

    To avoid errors due to missing CheckpointHandlers, concrete
    CheckpointManager constructor must allow mapping from item names to
    respective CheckpointHandlers to be input other than via save() and
    restore(). Please note that save() and restore() calls automatically
    map CheckpointHandlers to respective item names and retain it during the
    lifetime of the CheckpointManager instance.

    Example::

      # Single item
      mngr = ocp.CheckpointManager(directory)
      # No calls to save() or restore() before calling item_metadata().
      mngr.item_metadata(step)  # Raises error.

      mngr = ocp.CheckpointManager(directory,
          item_handlers=ocp.StandardCheckpointHandler)
      # No calls to save() or restore() before calling item_metadata().
      metadata = mngr.item_metadata(step)  # Successful.

      # Multiple items
      mngr = ocp.CheckpointManager(directory, item_names=('state', 'extra'))
      # No calls to save() or restore() before calling item_metadata().
      mngr.item_metadata(step)  # Raises error.

      mngr = ocp.CheckpointManager(directory,
        item_names=('state', 'extra'),
        item_handlers={
            'state': ocp.StandardCheckpointHandler,
            'extra': ocp.PytreeCheckpointHandler,
        }
      )
      # No calls to save() or restore() before calling item_metadata().
      metadata = mngr.item_metadata(step)  # Successful.

    Metadata may be None for an individual item.

    Args:
      step: Step for which to retrieve metadata.

    Returns:
      A dictionary mapping name to item metadata, or a single item metadata.
    """

  @abc.abstractmethod
  def metadata(self) -> Mapping[str, Any]:
    """Returns CheckpointManager level metadata if present, empty otherwise."""

  @abc.abstractmethod
  def metrics(self, step: int) -> Optional[PyTree]:
    """Returns metrics for step, if present."""

  @abc.abstractmethod
  def wait_until_finished(self):
    """Blocks until any incomplete save operations are completed.

    Note that this method will typically be a no-op if all checkpointers are
    synchronous, since old checkpoints are already cleaned up immediately after
    completing `save`, and there is no background thread to wait for.

    If some checkpointers are of type AsyncCheckpointer, however, this method
    will wait until each of these checkpointers is finished.
    """

  @abc.abstractmethod
  def check_for_errors(self):
    """Checks for any outstanding errors in completed asynchronous save operations.

    Delegates to underlying Checkpointer.
    """

  @abc.abstractmethod
  def close(self):
    """Waits for outstanding operations to finish and closes Checkpointers."""



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides async variants of path functions."""

from typing import Any

from etils import epath
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.path import step as step_lib


# TODO(b/360190539): This functionality should be provided by either an external
# library or Orbax should subclass epath.Path.
def async_makedirs(
    path: epath.Path,
    *args,
    parents: bool = False,
    exist_ok: bool = True,
    **kwargs,
):
  return asyncio_utils.as_async_function(path.mkdir)(
      *args, parents=parents, exist_ok=exist_ok, **kwargs
  )


def async_write_bytes(path: epath.Path, data: Any):
  return asyncio_utils.as_async_function(path.write_bytes)(data)


def async_exists(path: epath.Path):
  return asyncio_utils.as_async_function(path.exists)()


def async_stat(path: epath.Path):
  return asyncio_utils.as_async_function(path.stat)()


def async_rmtree(path: epath.Path):
  return asyncio_utils.as_async_function(path.rmtree)()


def async_is_tmp_checkpoint(path: epath.Path):
  return asyncio_utils.as_async_function(step_lib.is_tmp_checkpoint)(path)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""test cases for colossus snapshot."""

import os
from unittest import mock

from absl.testing import absltest
from etils import epath
from orbax.checkpoint._src.path.snapshot import snapshot



class DefaultSnapshotTest(absltest.TestCase):

  def __init__(self, *args, **kwargs):
    super(DefaultSnapshotTest, self).__init__(*args, **kwargs)
    self.source_dir = epath.Path(self.create_tempdir(name='source').full_path)
    # self.source_dir = '/tmp/test/path/to/source/'
    os.makedirs(self.source_dir, exist_ok=True)
    with open(os.path.join(self.source_dir, 'data.txt'), 'w') as f:
      f.write('data')

  def test_create_snapshot(self):
    dst_dir = '/tmp/test/path/to/dest/'

    # dst_dir = epath.Path(self.create_tempdir(name='dest').full_path)
    self.assertFalse(os.path.exists(dst_dir))
    snapshot.DefaultSnapshot.create_snapshot(str(self.source_dir), str(dst_dir))
    self.assertTrue(os.path.exists(dst_dir))
    with open(os.path.join(dst_dir, 'data.txt')) as f:
      self.assertEqual('data', f.read())
      f.close()

  def test_release_snapshot(self):
    self.assertTrue(os.path.exists(self.source_dir))
    self.assertTrue(snapshot.DefaultSnapshot.release_snapshot(self.source_dir))
    self.assertFalse(os.path.exists(self.source_dir))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for JsonCheckpointHandler."""

import time
from typing import Optional

from absl import flags
from absl.testing import absltest
from etils import epath
import jax
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.handlers import json_checkpoint_handler

# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

FLAGS = flags.FLAGS

JsonCheckpointHandler = json_checkpoint_handler.JsonCheckpointHandler
JsonSaveArgs = json_checkpoint_handler.JsonSaveArgs
JsonRestoreArgs = json_checkpoint_handler.JsonRestoreArgs


class TestJsonCheckpointHandler(JsonCheckpointHandler):

  def __init__(
      self,
      filename: Optional[str] = None,
  ):
    super().__init__(filename=filename)
    self._handler = json_checkpoint_handler.JsonCheckpointHandler
    self._filename = filename or 'metadata'

  def _save_fn(self, x, directory):
    time.sleep(5)
    return super()._save_fn(x, directory)

  def close(self):
    self._executor.shutdown(wait=True)


class JsonCheckpointHandlerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.directory = epath.Path(
        self.create_tempdir(name='checkpointing_test').full_path
    )

  def test_save_restore(self):
    item = {'a': 1, 'b': {'c': 'test1', 'b': 'test2'}, 'd': 5.5}
    handler = JsonCheckpointHandler()
    handler.save(self.directory, args=JsonSaveArgs(item))
    restored = handler.restore(self.directory)
    self.assertEqual(item, restored)

  def test_save_restore_filename(self):
    item = {'a': 1, 'b': {'c': 'test1', 'b': 'test2'}, 'd': 5.5}
    handler = JsonCheckpointHandler(filename='file')
    handler.save(self.directory, args=JsonSaveArgs(item))
    restored = handler.restore(self.directory)
    self.assertEqual(item, restored)
    self.assertTrue((self.directory / 'file').exists())
    self.assertFalse((self.directory / 'metadata').exists())

  def test_async_save_restore(self):
    item = {'a': 1, 'b': {'c': 'test1', 'b': 'test2'}, 'd': 5.5}
    handler = TestJsonCheckpointHandler()

    async def run_async_test():
      async_futures = await handler.async_save(
          self.directory, args=JsonSaveArgs(item)
      )
      # item wasn't saved at directory right away.
      self.assertFalse((self.directory / 'metadata').exists())
      for future in async_futures:
        future.result()
      self.assertTrue((self.directory / 'metadata').exists())
      restored = handler.restore(directory=self.directory, args=None)
      self.assertEqual(item, restored)

    asyncio_utils.run_sync(run_async_test())
    handler.close()


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at step level."""

import dataclasses
from typing import Any

from absl import logging
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
StepMetadata = checkpoint.StepMetadata
ItemMetadata = checkpoint.ItemMetadata
StepStatistics = step_statistics.SaveStepStatistics


def serialize(metadata: StepMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""

  # Part of the StepMetadata api for user convenience, but not saved to disk.
  if metadata.item_metadata is not None:
    just_item_names = {k: None for k in metadata.item_metadata.keys()}
  else:
    just_item_names = None

  # Save only float performance metrics.
  performance_metrics = metadata.performance_metrics
  float_metrics = {
      metric: val
      for metric, val in dataclasses.asdict(performance_metrics).items()
      if isinstance(val, float)
  }

  return {
      'format': metadata.format,
      'item_handlers': metadata.item_handlers,
      'item_metadata': just_item_names,
      'metrics': metadata.metrics,
      'performance_metrics': float_metrics,
      'init_timestamp_nsecs': metadata.init_timestamp_nsecs,
      'commit_timestamp_nsecs': metadata.commit_timestamp_nsecs,
      'custom': metadata.custom,
  }


def deserialize(
    metadata_dict: SerializedMetadata,
    item_metadata: ItemMetadata | None = None,
    metrics: dict[str, Any] | None = None,
) -> StepMetadata:
  """Deserializes `metadata_dict` and other kwargs to `StepMetadata`."""
  validated_metadata_dict = {}

  utils.validate_field(metadata_dict, 'format', str)
  validated_metadata_dict['format'] = metadata_dict.get('format', None)

  utils.validate_field(metadata_dict, 'item_handlers', dict)
  for k in metadata_dict.get('item_handlers', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'item_handlers', k, str, str)
  validated_metadata_dict['item_handlers'] = metadata_dict.get(
      'item_handlers', {}
  )

  utils.validate_field(metadata_dict, 'item_metadata', dict)
  for k in metadata_dict.get('item_metadata', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'item_metadata', k, str)
  validated_metadata_dict['item_metadata'] = metadata_dict.get(
      'item_metadata', None
  )
  if item_metadata is not None:
    if validated_metadata_dict['item_metadata'] is None:
      validated_metadata_dict['item_metadata'] = {}
    utils.validate_field(item_metadata, 'item_metadata', dict)
    for k, v in item_metadata.items():
      utils.validate_dict_entry(metadata_dict, 'item_metadata', k, str)
      validated_metadata_dict['item_metadata'][k] = v

  utils.validate_field(metadata_dict, 'metrics', dict)
  for k in metadata_dict.get('metrics', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'metrics', k, str)
  validated_metadata_dict['metrics'] = metadata_dict.get('metrics', {})
  if metrics is not None:
    utils.validate_field(metadata_dict, 'metrics', dict)
    for k, v in metrics.items():
      utils.validate_dict_entry(metadata_dict, 'metrics', k, str)
      validated_metadata_dict['metrics'][k] = v

  utils.validate_field(metadata_dict, 'performance_metrics', dict)
  for k in metadata_dict.get('performance_metrics', {}) or {}:
    utils.validate_dict_entry(
        metadata_dict, 'performance_metrics', k, str, float
    )
  validated_metadata_dict['performance_metrics'] = StepStatistics(
      **metadata_dict.get('performance_metrics', {})
  )

  utils.validate_field(metadata_dict, 'init_timestamp_nsecs', int)
  validated_metadata_dict['init_timestamp_nsecs'] = (
      metadata_dict.get('init_timestamp_nsecs', None)
  )

  utils.validate_field(metadata_dict, 'commit_timestamp_nsecs', int)
  validated_metadata_dict['commit_timestamp_nsecs'] = (
      metadata_dict.get('commit_timestamp_nsecs', None)
  )

  utils.validate_field(metadata_dict, 'custom', dict)
  for k in metadata_dict.get('custom', {}) or {}:
    utils.validate_dict_entry(metadata_dict, 'custom', k, str)
  validated_metadata_dict['custom'] = metadata_dict.get('custom', {})

  for k in metadata_dict:
    if k not in validated_metadata_dict:
      logging.warning(
          'Provided metadata contains unknown key %s, ignoring.', k
      )

  return StepMetadata(**validated_metadata_dict)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for empty values as leafs in the checkpoint tree."""

from absl.testing import absltest
from absl.testing import parameterized
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.testing import test_tree_utils


class EmptyValuesTest(parameterized.TestCase):

  @parameterized.parameters(
      (1, False),
      (dict(), True),
      ({}, True),
      ({"a": {}}, False),
      ([], True),
      ([[]], False),
      (None, True),
      ((1, 2), False),
      (test_tree_utils.EmptyNamedTuple(), False),
      (test_tree_utils.MuNu(mu=None, nu=None), False),
      (test_tree_utils.NamedTupleWithNestedAttributes(), False),
      (test_tree_utils.NamedTupleWithNestedAttributes(nested_dict={}), False),
  )
  def test_is_supported_empty_value(self, value, expected):
    self.assertEqual(expected, empty_values.is_supported_empty_value(value))


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AsyncCheckpointHandler interface."""

import abc
from typing import List, Optional

from etils import epath
from orbax.checkpoint import future
from orbax.checkpoint._src.handlers import checkpoint_handler


class AsyncCheckpointHandler(checkpoint_handler.CheckpointHandler):
  """An interface providing async methods that can be used with CheckpointHandler."""

  @abc.abstractmethod
  async def async_save(
      self, directory: epath.Path, *args, **kwargs
  ) -> Optional[List[future.Future]]:
    """Constructs a save operation.

    Synchronously awaits a copy of the item, before returning commit futures
    necessary to save the item.

    Args:
      directory: the directory to save to.
      *args: additional arguments for save.
      **kwargs: additional arguments for save.
    """
    pass



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for utils.py."""
from absl.testing import absltest
from absl.testing import parameterized
from orbax.checkpoint import msgpack_utils


class MsgpackUtilsTest(parameterized.TestCase):

  @parameterized.parameters(
      (3,),
      ('foo',),
      ((1,),),
      (('a', 1, 1.2),),
      ([],),
      ({'a': [2, 'hi', (1, 3, 'bye')], 'b': {'hi': 2}},),
  )
  def test_serialize(self, x):
    self.assertEqual(
        x, msgpack_utils.msgpack_restore(msgpack_utils.msgpack_serialize(x))
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Counters used to make barrier names unique.

Do not add any functions unless they conform precisely to the existing pattern.
"""

import itertools


_async_save_counter = itertools.count()
_composite_save_counter = itertools.count()
_tmp_directory_counter = itertools.count()
_model_surgery_loading_counter = itertools.count()
_root_metadata_save_counter = itertools.count()

_global_max_broadcast_counter = itertools.count()
_local_all_steps_broadcast_counter = itertools.count()
_find_complete_slice_broadcast_counter = itertools.count()


def async_save_counter() -> str:
  return str(next(_async_save_counter))


def composite_save_counter() -> str:
  return str(next(_composite_save_counter))


def tmp_directory_counter() -> str:
  return str(next(_tmp_directory_counter))


def model_surgery_loading_counter() -> str:
  return str(next(_model_surgery_loading_counter))


def root_metadata_save_counter() -> str:
  return str(next(_root_metadata_save_counter))


# Emergency checkpointing counters.


def global_max_broadcast_counter() -> str:
  return str(next(_global_max_broadcast_counter))


def local_all_steps_broadcast_counter() -> str:
  return str(next(_local_all_steps_broadcast_counter))


def find_complete_slice_broadcast_counter() -> str:
  return str(next(_find_complete_slice_broadcast_counter))



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
from orbax.checkpoint._src.arrays import numpy_utils as np_utils


class PrettySliceTest(parameterized.TestCase):

  @parameterized.parameters([
      (slice(None), ':'),
      (slice(4), ':4'),
      (slice(1, 4), '1:4'),
      (slice(1, 4, 2), '1:4:2'),
  ])
  def test_produces_extended_slice_syntax(self, s: slice, s_string: str):
    self.assertEqual(s_string, np_utils._pretty_slice(s))
    self.assertEqual(s, eval(f'np.s_[{s_string}]'))  # pylint:disable=eval-used


class ResolveDissolveSliceTest(parameterized.TestCase):

  @parameterized.parameters([
      ((3, 4, 5), np.s_[:, :, :], np.s_[0:3:1, 0:4:1, 0:5:1]),
      ((3, 4, 5), np.s_[:2, 1:, ::2], np.s_[0:2:1, 1:4:1, 0:5:2]),
      ((), np.s_[...], np.s_[()]),
  ])
  def test_round_trip(
      self,
      array_shape: np_utils.Shape,
      implicit_slice: np_utils.NdSlice,
      explicit_slice: np_utils.NdSlice,
  ):
    self.assertEqual(
        explicit_slice,
        np_utils.resolve_slice(implicit_slice, array_shape),
    )
    self.assertEqual(
        implicit_slice,
        np_utils.dissolve_slice(explicit_slice, array_shape),
    )

  def test_dissolve_slice_can_remove_redundant_rank(self):
    self.assertEqual(
        (np.s_[1:2],),
        np_utils.dissolve_slice(
            np.s_[1:2, 0:4, 0:5:1], (3, 4, 5), preserve_rank=False
        ),
    )
    self.assertEqual(
        np.s_[:, 1:2],
        np_utils.dissolve_slice(
            np.s_[0:3:1, 1:2:1, 0:5:1], (3, 4, 5), preserve_rank=False
        ),
    )
    self.assertEqual(
        ...,
        np_utils.dissolve_slice(
            np.s_[:3, 0:4, 0:5:1], (3, 4, 5), preserve_rank=False
        ),
    )


class PrettyNdSliceTest(parameterized.TestCase):

  @parameterized.parameters([
      (..., 'np.s_[...]'),
      ((slice(1, 4, 2), slice(None)), 'np.s_[1:4:2, :]'),
  ])
  def test_generally_produces_np_s_syntax(
      self,
      idx: np_utils.Index,
      idx_string: str,
  ):
    self.assertEqual(idx_string, np_utils.pretty_nd_slice(idx))
    self.assertEqual(idx, eval(idx_string))  # pylint:disable=eval-used

  def test_rank_1_slice(self):
    # A tuple of length 1 won't exactly round-trip, but it will produce
    # a value that can be used in the same places, because `a[(s,)]`
    # will do the same as `a[s]`.
    idx = (slice(1, 4, 2),)
    self.assertEqual('np.s_[1:4:2]', np_utils.pretty_nd_slice(idx))

  def test_empty_tuple_is_rendered_as_ellipsis(self):
    # Similarly, an empty won't exactly round-trip, but it will produce
    # a value that can be used in the same places, because `a[...]`
    # will do the same as `a[()]` (unless `a` has rank 0 in which case
    # ellipsis actually does the more useful thing).
    self.assertEqual('np.s_[...]', np_utils.pretty_nd_slice(()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Management of fragments of arrays.

A fragment is a lot like a shard but its shape is not constrained by any
relationship to a mesh of devices, or to other fragments.
"""

import dataclasses
from typing import Sequence

import jax
import numpy as np
from orbax.checkpoint._src.arrays import numpy_utils as np_utils
from orbax.checkpoint._src.arrays import types

Shape = types.Shape
Index = types.Index


def _ndarray_from_index(idx: Index) -> np.ndarray:
  if idx:
    return np.stack([np_utils.int_tuple_from_slice(s) for s in idx])
  else:
    return np.empty([0, 3], dtype=int)


def _index_from_ndarray(a: np.ndarray) -> Index:
  return tuple(slice(*xs) for xs in a)


@dataclasses.dataclass(frozen=True, init=False)
class Fragment:
  """One of a collection of slices into the same (abstract or concrete) array.

  Fields:
    np_index: A (start, stop, step) triple for each dimension, that locates
      the fragment within the global shape. A sequence of `slice` objects can be
      used to express this as well (see the `index` parameter to `__init__`) but
      this form is more convenient because it can't contain `None`.
    value: The data for this fragment. If this is `None`, the fragment is
      abstract.
  """

  np_index: np.ndarray  # shape=[{rank}, 3], dtype=int
  value: np.ndarray | None = None

  def __init__(
      self,
      *,
      index: Index | None = None,
      np_index: np.ndarray | None = None,
      value: np.ndarray | None = None,
  ):
    if value is not None and not isinstance(value, np.ndarray):
      raise TypeError(
          f'Fragment value must be an np.ndarray (or None), not {type(value)}.'
      )

    if index is not None and np_index is not None:
      raise ValueError('Cannot specify both index and np_index.')
    if index is None and np_index is None:
      raise ValueError('Must specify either index or np_index.')
    if index is not None:
      if not isinstance(index, tuple):
        raise TypeError(
            f'Fragment index must be a tuple of slices, got {type(index)}.'
        )
      np_index = _ndarray_from_index(index)
    elif not isinstance(np_index, np.ndarray):
      raise TypeError(
          f'Fragment np_index must be an np.ndarray , got {type(np_index)}.'
      )

    object.__setattr__(self, 'value', value)
    object.__setattr__(self, 'np_index', np_index)

  @property
  def index(self) -> Index:
    """A tuple of slices, locating this fragment in the global array."""
    return _index_from_ndarray(self.np_index)

  @property
  def start(self) -> np.ndarray:  # shape=[{rank}]
    return self.np_index[:, 0]

  @property
  def stop(self) -> np.ndarray:  # shape=[{rank}]
    return self.np_index[:, 1]

  @property
  def step(self) -> np.ndarray:  # shape=[{rank}]
    return self.np_index[:, 2]

  @property
  def shape(self) -> Shape:
    return tuple(np_utils.slice_shape(self.index))

  @property
  def size(self) -> int:
    return np.prod(self.shape)

  def __eq__(self, other: 'Fragment'):
    if not isinstance(other, Fragment):
      return False
    if not np.array_equal(self.np_index, other.np_index):
      return False
    self_value = self.value
    other_value = other.value
    if self_value is None:
      return other_value is None
    else:
      return (
          other_value is not None
          and self_value.dtype == other_value.dtype
          and np.array_equal(self_value, other_value)
      )

  def __repr__(self):
    maybe_value_str = ', value=...' if self.value is not None else ''
    return (
        f'Fragment(index={np_utils.pretty_nd_slice(self.index)}'
        f'{maybe_value_str})'
    )

  def __array__(self) -> np.ndarray:
    if self.value is None:
      raise ValueError(f"Can't convert abstract fragment to array: {self!r}.'")
    return self.value  # pytype: disable=bad-return-type

  def is_degenerate(self) -> bool:
    """Whether the index has any slices of length zero."""
    return (self.stop == self.start).any()

  @property
  def nbytes(self) -> int:
    if self.value is None:
      raise ValueError(
          "Can't compute nbytes of abstract fragment. Use nbytes_astype()."
      )
    return self.value.nbytes

  def nbytes_astype(self, dtype: np.dtype) -> int:
    return np.prod([dtype.itemsize, *self.shape])

  def offset_by(self, delta: np.ndarray) -> 'Fragment':
    out_idx = self.np_index.copy()
    out_idx[:, :2] += np.expand_dims(delta, axis=1)
    return Fragment(np_index=out_idx, value=self.value)


@dataclasses.dataclass(frozen=True)
class Fragments:
  """An abstract or concrete collection of fragments.

  A `Fragments` is a lot like a `jax.Array` (or a `jax.ShapeDtypeStruct`) but
  there is a weaker relationship between the indices of the `Fragment`
  instances that it carries than there is between the indices of the shards
  of a `jax.Array` (fragments are not required to have the same shape, or to map
  to a device mesh).
  """

  shape: Shape
  dtype: np.dtype
  fragments: Sequence[Fragment]

  def __post_init__(self):
    for fragment in self.fragments:
      if not isinstance(fragment, Fragment):
        raise TypeError(
            f'Fragments must contain Fragment, not {type(fragment)}.'
        )

  def is_degenerate(self) -> bool:
    """Whether this contains only degenerate fragments."""
    return all(f.is_degenerate() for f in self.fragments)

  @property
  def nbytes(self) -> int:
    """The total number of bytes for the global shape of this object."""
    return np.prod((self.dtype.itemsize, *self.shape))

  @property
  def addressable_nbytes(self) -> int:
    """The total number of bytes for the fragments collected in this object."""
    return sum(f.nbytes_astype(self.dtype) for f in self.fragments)

  def __array__(self) -> np.ndarray:
    for f in self.fragments:
      if f.value is None:
        raise ValueError(
            f"Can't convert abstract fragments to array: {self!r}.'"
        )
    non_degenerate_fragments = [
        f for f in self.fragments if not f.is_degenerate()
    ]
    for f in non_degenerate_fragments:
      # Fast path, avoiding creating a copy if we just have a single fragment
      # which has all the data.
      # TODO(b/330745907): we can relax/rewrite this check to search for at
      # least one fragment that covers the target shape fully (omitting the
      # fragments that have step > 1 on any dimension).
      if f.shape == self.shape:
        return f.value  # pytype: disable=bad-return-type
    if not _is_full(self):
      raise ValueError(
          f'Attempt to convert non-full Fragments to array: {self}.'
      )
    result = np.empty(self.shape, dtype=self.dtype)
    for f in non_degenerate_fragments:
      result[f.index] = f.value
    return result


def _is_full(fragments: Fragments) -> bool:
  """True iff every array element is covered by some fragment."""
  present = np.zeros(fragments.shape, dtype=bool)
  for f in fragments.fragments:
    present[f.index] = True
  return np.all(present)


def _normalize(idx: Index, shape: Shape) -> Index:
  return tuple(
      slice(
          s.start if s.start is not None else 0,
          s.stop if s.stop is not None else dim,
          s.step if s.step is not None else 1,
      )
      for s, dim in zip(idx, shape)
  )


def addressable_shards(x: jax.Array | jax.ShapeDtypeStruct) -> list[Index]:
  """Computes list of fragment indices for addressable shards of a JAX array."""
  sharding = getattr(x, 'sharding', None)
  shape = x.shape
  if not sharding:
    return [tuple(slice(0, dim, 1) for dim in shape)]
  return [
      _normalize(idx, shape)
      for idx in sharding.addressable_devices_indices_map(shape).values()
  ]


def abstract_fragments(
    x: jax.Array | jax.ShapeDtypeStruct | Fragments,
) -> Fragments:
  """Returns abstract fragments matching the given array."""
  if isinstance(x, Fragments):
    return x
  return Fragments(
      x.shape,
      x.dtype,
      [Fragment(index=index, value=None) for index in addressable_shards(x)],
  )


def validate_fragments_can_be_stacked(fragments: Fragments) -> None:
  """Validates that the given fragments can be stacked."""
  if not fragments.fragments:
    raise ValueError('No fragments to stack.')
  fragment_arrays = [
      fragment.value
      for fragment in fragments.fragments
      if fragment.value is not None
  ]
  if len(fragment_arrays) != len(fragments.fragments):
    raise ValueError(f'Not all fragments have values: {fragments}')
  fragment_shape = fragment_arrays[0].shape
  if not all(a.shape == fragment_shape for a in fragment_arrays):
    raise ValueError(
        f"Differently-shaped fragments can't be cast to shards: {fragments}"
    )


def stack_fragments(fragments: Fragments | None) -> np.ndarray | None:
  """Stacks the given fragments, which must all have the same shape."""
  if fragments is None:
    return fragments
  validate_fragments_can_be_stacked(fragments)
  fragment_arrays = [fragment.value for fragment in fragments.fragments]
  return (
      np.expand_dims(fragment_arrays[0], axis=0)
      if len(fragment_arrays) == 1
      else np.stack(fragment_arrays)
  )



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import math
import os
import pathlib
import tracemalloc as tm
from typing import Any
import unittest

from absl.testing import absltest
from absl.testing import parameterized
import jax
from jax import dtypes as _dtypes
from jax.experimental import layout
import jax.numpy as jnp
import numpy as np
from orbax.checkpoint import future
from orbax.checkpoint import test_utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
import tensorstore as ts


GSPMDSharding = jax.sharding.GSPMDSharding
NamedSharding = jax.sharding.NamedSharding
P = jax.sharding.PartitionSpec
DLL = layout.DeviceLocalLayout
Layout = layout.Layout

jax.config.update('jax_enable_x64', True)


def _dtype(x):
  if hasattr(x, 'dtype'):
    return x.dtype
  elif type(x) in _dtypes.python_scalar_dtypes:
    return np.dtype(_dtypes.python_scalar_dtypes[type(x)])
  else:
    return np.asarray(x).dtype


def serialize(arrs, tspecs):
  async def _serialize():
    await asyncio.gather(*[
        serialization.async_serialize(arr, tspec)
        for arr, tspec in zip(arrs, tspecs)
    ])

  asyncio_utils.run_sync(_serialize())
  test_utils.sync_global_processes('serialization_complete')


def deserialize(
    shardings, tensorstore_specs, global_shapes=None, dtypes=None, strict=True
):
  if global_shapes is None:
    global_shapes = [None for _ in tensorstore_specs]
  if dtypes is None:
    dtypes = [None for _ in tensorstore_specs]

  async def _deserialize():
    return await asyncio.gather(*[
        serialization.async_deserialize(
            sharding, tspec, shape, dtype, strict=strict
        )
        for sharding, tspec, shape, dtype in zip(
            shardings, tensorstore_specs, global_shapes, dtypes
        )
    ])

  result = asyncio_utils.run_sync(_deserialize())
  test_utils.sync_global_processes('deserialization_complete')
  return result


class FutureWithSpeedbump(future.Future):

  def __init__(self, f, speedbump):
    self._f = f
    self._speedbump = speedbump
    assert self._speedbump >= 0

  def result(self, timeout: int | None = None) -> Any:
    raise NotImplementedError()

  async def _sleep_and_result(self):
    await asyncio.sleep(self._speedbump)
    return await self._f

  def __await__(self):
    return self._sleep_and_result().__await__()


def create_global_mesh(mesh_shape, axis_names):
  size = math.prod(mesh_shape)
  if len(jax.devices()) < size:
    raise unittest.SkipTest(f'Test requires {size} global devices.')
  devices = sorted(jax.devices(), key=lambda d: d.id)
  mesh_devices = np.array(devices[:size]).reshape(mesh_shape)
  global_mesh = jax.sharding.Mesh(mesh_devices, axis_names)
  return global_mesh


class CheckpointTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.ckpt_dir = pathlib.Path(self.create_tempdir('ckpt').full_path)
    test_utils.sync_global_processes('CheckpointTest:setup_complete')

  def tearDown(self):
    test_utils.sync_global_processes('CheckpointTest:tests_complete')
    super().tearDown()

  def assertArraysEqual(
      self,
      x,
      y,
      *,
      check_dtypes=True,
      err_msg='',
      allow_object_dtype=False,
      verbose=True,
  ):
    """Assert that x and y arrays are exactly equal."""
    if check_dtypes:
      self.assertDtypesMatch(x, y)
    x = np.asarray(x)
    y = np.asarray(y)

    if (not allow_object_dtype) and (x.dtype == object or y.dtype == object):
      # See https://github.com/google/jax/issues/17867
      raise TypeError(
          'assertArraysEqual may be poorly behaved when np.asarray casts to'
          ' dtype=object. If comparing PRNG keys, consider'
          ' random_test.KeyArrayTest.assertKeysEqual. If comparing collections'
          ' of arrays, consider using assertAllClose. To let this test proceed'
          ' anyway, pass allow_object_dtype=True.'
      )

    # Work around https://github.com/numpy/numpy/issues/18992
    with np.errstate(over='ignore'):
      np.testing.assert_array_equal(x, y, err_msg=err_msg, verbose=verbose)

  def assertDtypesMatch(self, x, y):
    self.assertEqual(_dtype(x), _dtype(y))

  def test_memory_consumption(self):
    global_mesh = create_global_mesh((2, 4), ('x', 'y'))
    inp_shape = (2_048, 4_096)
    pspec = P('x', 'y')
    num = math.prod(inp_shape)
    sharding = NamedSharding(global_mesh, pspec)
    src = jnp.arange(num, dtype=np.int32).reshape(inp_shape)  # 8e9
    inp = jax.make_array_from_callback(
        inp_shape, sharding, lambda idx: src[idx]
    )
    tspec = serialization.get_tensorstore_spec(str(self.ckpt_dir))

    serialize(
        [inp],
        [tspec],
    )

    async def deserialize_with_byte_limit():
      r = await serialization.async_deserialize(
          sharding,
          tspec,
          inp_shape,
          byte_limiter=serialization.LimitInFlightBytes(4_200_000),
      )
      r.block_until_ready()

    tm.start()
    asyncio_utils.run_sync(deserialize_with_byte_limit())
    unused_current, peak = tm.get_traced_memory()
    # NB: some padding + tensorstore overhead. It should always be
    # less than array size (2048 * 4096 * 4 = 32M)
    self.assertLess(peak, 10_000_000)
    deserialize_wo_limit = serialization.async_deserialize(
        sharding, tspec, inp_shape
    )
    tm.clear_traces()
    # NB: call block_until_ready() is important here and above
    # because otherwise this leads to racing condition and segfault with
    # tensorstore attempting to dealloc using tracemalloc which is already
    # destroyed.
    asyncio_utils.run_sync(deserialize_wo_limit).block_until_ready()

    unused_current, peak = tm.get_traced_memory()
    # We load entire array in memory here.
    self.assertGreater(peak, 30_000_000)
    tm.stop()

  def test_checkpointing_jax_array(self):
    global_mesh = create_global_mesh((4, 2), ('x', 'y'))
    inp_shape = (8, 2)
    pspec = P('x', 'y')
    num = math.prod(inp_shape)

    # First Array
    global_input_data1 = np.arange(num, dtype=np.int32).reshape(inp_shape)
    a1 = jax.make_array_from_callback(
        inp_shape,
        NamedSharding(global_mesh, pspec),
        lambda idx: global_input_data1[idx],
    )

    # Second Array
    global_input_data2 = np.arange(num, num + num, dtype=np.int32).reshape(
        inp_shape
    )
    a2 = jax.make_array_from_callback(
        inp_shape,
        NamedSharding(global_mesh, pspec),
        lambda idx: global_input_data2[idx],
    )

    # Third Array
    def cb3(_):
      return np.array([], dtype=np.float32)

    global_mesh1d = create_global_mesh((8,), ('x',))
    a3 = jax.make_array_from_callback(
        (0,), NamedSharding(global_mesh1d, P(None)), cb3
    )

    ckpt_paths = [
        self.create_tempdir(f'{self.ckpt_dir}/{i}').full_path for i in range(3)
    ]
    test_utils.sync_global_processes(
        'test_checkpointing_jax_array:create_arr_paths'
    )
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([a1, a2, a3], tspecs)

    m1, m2, m3 = deserialize(
        [
            NamedSharding(global_mesh, pspec),
            NamedSharding(global_mesh, P('x')),
            NamedSharding(global_mesh1d, P(None)),
        ],
        tspecs,
    )

    logging.info(m1.addressable_shards)
    logging.info(m2.addressable_shards)
    logging.info(m3.addressable_shards)
    self.assertIsInstance(m1, jax.Array)
    self.assertArraysEqual(
        np.asarray(m1.addressable_shards[0].data),
        np.array([[0], [2]], dtype=np.int32),
    )
    self.assertArraysEqual(
        np.asarray(m1.addressable_shards[1].data),
        np.array([[1], [3]], dtype=np.int32),
    )
    self.assertEqual(m1.addressable_shards[0].data.shape, (2, 1))
    self.assertEqual(m1.dtype, np.int32)

    self.assertIsInstance(m2, jax.Array)
    self.assertArraysEqual(
        np.asarray(m2.addressable_shards[0].data),
        np.array([[16, 17], [18, 19]], dtype=np.int32),
    )
    self.assertArraysEqual(
        np.asarray(m2.addressable_shards[1].data),
        np.array([[16, 17], [18, 19]], dtype=np.int32),
    )
    self.assertEqual(m2.addressable_shards[0].data.shape, (2, 2))
    self.assertEqual(m2.dtype, np.int32)

    self.assertIsInstance(m3, jax.Array)
    for i, s in enumerate(m3.addressable_shards):
      self.assertEqual(s.index, (slice(None),))
      self.assertEqual(s.replica_id, i)
      self.assertArraysEqual(np.asarray(s.data), np.array([], dtype=np.float32))
    self.assertEqual(m3.dtype, np.float32)

  @parameterized.product(input_dtype=[np.int32, jnp.bfloat16])
  def test_checkpointing_with_bigger_shape_jax_array(self, input_dtype):
    global_mesh = create_global_mesh((2, 2), ('x', 'y'))
    global_input_shape = (8, 2)
    num = math.prod(global_input_shape)

    global_input_data1 = np.arange(num, dtype=input_dtype).reshape(
        global_input_shape
    )

    def cb1(index):
      return global_input_data1[index]

    arr = jax.make_array_from_callback(
        global_input_shape, NamedSharding(global_mesh, P('x', 'y')), cb1
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([arr], tspecs)

    ds = NamedSharding(create_global_mesh((4, 2), ('x', 'y')), P('x', 'y'))

    (m1,) = deserialize([ds], tspecs, [(12, 2)], [np.float32], strict=False)

    expected_data = {
        0: np.array([[0], [2], [4]], dtype=np.float32),
        1: np.array([[1], [3], [5]], dtype=np.float32),
        2: np.array([[6], [8], [10]], dtype=np.float32),
        3: np.array([[7], [9], [11]], dtype=np.float32),
        4: np.array([[12], [14], [0]], dtype=np.float32),
        5: np.array([[13], [15], [0]], dtype=np.float32),
        6: np.array([[0], [0], [0]], dtype=np.float32),
        7: np.array([[0], [0], [0]], dtype=np.float32),
    }

    for l in m1.addressable_shards:
      self.assertArraysEqual(np.asarray(l.data), expected_data[l.device.id])

    new_ds = GSPMDSharding.get_replicated(list(global_mesh.devices.flat))
    (m2,) = deserialize([new_ds], tspecs, [(8, 2)], [np.float32])
    for l in m2.addressable_shards:
      self.assertArraysEqual(l.data, global_input_data1.astype('float32'))

  @parameterized.product(input_dtype=[jnp.int4, jnp.int8])
  def test_checkpointing_with_int4(self, input_dtype):
    global_mesh = create_global_mesh((2, 2), ('x', 'y'))
    global_input_shape = (8, 2)
    num = math.prod(global_input_shape)

    global_input_data = np.arange(num, dtype=input_dtype).reshape(
        global_input_shape
    )

    def cb(index):
      return global_input_data[index]

    arr = jax.make_array_from_callback(
        global_input_shape, NamedSharding(global_mesh, P('x', 'y')), cb
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([arr], tspecs)

    ds = NamedSharding(create_global_mesh((4, 2), ('x', 'y')), P('x', 'y'))

    target_dtype = jnp.dtype('int4')
    (m1,) = deserialize([ds], tspecs, [(12, 2)], [target_dtype], strict=False)

    # values bigger than 7 are converted properly.
    expected_data = {
        0: jnp.array([[0], [2], [4]], dtype=target_dtype),
        1: jnp.array([[1], [3], [5]], dtype=target_dtype),
        2: jnp.array([[6], [8], [10]], dtype=target_dtype),
        3: jnp.array([[7], [9], [11]], dtype=target_dtype),
        4: jnp.array([[12], [14], [0]], dtype=target_dtype),
        5: jnp.array([[13], [15], [0]], dtype=target_dtype),
        6: jnp.array([[0], [0], [0]], dtype=target_dtype),
        7: jnp.array([[0], [0], [0]], dtype=target_dtype),
    }

    for l in m1.addressable_shards:
      self.assertArraysEqual(np.asarray(l.data), expected_data[l.device.id])

    new_ds = GSPMDSharding.get_replicated(list(global_mesh.devices.flat))
    (m2,) = deserialize([new_ds], tspecs, [(8, 2)], [target_dtype])
    for l in m2.addressable_shards:
      self.assertArraysEqual(l.data, global_input_data.astype(target_dtype))

  def test_checkpointing_scalar_jax_array(self):
    global_mesh = create_global_mesh((2,), 'x')
    global_input_shape = ()
    data = np.array(4)
    s = NamedSharding(global_mesh, P(None))
    array1 = jax.make_array_from_callback(
        global_input_shape, s, lambda idx: data[idx]
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([array1], tspecs)

    ds = NamedSharding(global_mesh, P(None))

    (m1,) = deserialize([ds], tspecs, [()], [np.float32])

    for l in m1.addressable_shards:
      self.assertArraysEqual(np.asarray(l.data), data.astype(np.float32))

  def test_deserialize_tensorstore_array_jax_array(self):
    global_mesh = create_global_mesh((2,), 'x')
    data = np.arange(1024)
    tspec = ts.array(data).spec()
    (m1,) = deserialize([NamedSharding(global_mesh, P(None))], [tspec])
    for l in m1.addressable_shards:
      self.assertArraysEqual(np.asarray(l.data), data)

  def test_spec_has_metadata(self):
    spec = {
        'a': {
            'b': 1,
            'c': 2,
        },
        'd': 3,
        'e': {'a': 2, 'metadata': 3},
        'f': 4,
    }
    self.assertTrue(serialization._spec_has_metadata(spec))
    self.assertTrue(
        serialization._spec_has_metadata({
            'driver': 'zarr',
            'kvstore': 'gfile',
            'metadata': {'chunks': 4, 'shape': (32, 64)},
            'one_more': 'thing',
        })
    )

  def test_spec_has_no_metadata(self):
    spec = {
        'a': {
            'b': 1,
            'c': 2,
        },
        'd': 3,
        'e': {
            'a': 2,
        },
        'f': 4,
    }
    self.assertFalse(serialization._spec_has_metadata(spec))

  def test_empty_spec_has_no_metadata(self):
    spec = {}
    self.assertFalse(serialization._spec_has_metadata(spec))

  @parameterized.named_parameters(
      ('gcs', 'gs://my/ckpt/dir/path'), ('file', '/my/ckpt/dir/path')
  )
  def test_get_tensorstore_spec_ocdbt(self, path):
    spec = serialization.get_tensorstore_spec(path, ocdbt=True)
    is_gcs_path = path.startswith('gs://')
    if is_gcs_path:
      self.assertEqual(spec['kvstore']['base'], os.path.dirname(path))
    else:
      self.assertEqual(
          spec['kvstore']['base'],
          {
              'driver': ts_utils.DEFAULT_DRIVER,
              'path': os.path.dirname(path),
          },
      )
    self.assertEqual(spec['kvstore']['path'], 'path')

  def test_get_tensorstore_spec_not_absolute_path(self):
    path = 'my/ckpt/path'
    with self.assertRaisesRegex(
        ValueError, 'Checkpoint path should be absolute'
    ):
      serialization.get_tensorstore_spec(path, ocdbt=True)

  def test_maybe_cloud_storage(self):
    gs_path = 'gs://some-buck/path'
    gs_spec = serialization.get_tensorstore_spec(gs_path, ocdbt=True)
    self.assertTrue(serialization.is_remote_storage(gs_spec))

    local_path = '/tmp/checkpoint'
    local_spec = serialization.get_tensorstore_spec(local_path, ocdbt=True)
    self.assertFalse(serialization.is_remote_storage(local_spec))

    nested_tspec = {
        'driver': 'cast',
        'dtype': 'int32',
        'base': {
            'driver': 'zarr',
            'kvstore': {'driver': 'ocdbt', 'base': 's3://some-bucket/path'},
        },
    }
    self.assertTrue(serialization.is_remote_storage(nested_tspec))

  def test_deserialization_with_int4(self):
    dtype = jnp.int4
    shape = (8, 2)
    arr = jnp.arange(np.prod(shape)).reshape(shape).astype(dtype)

    # Run serialization.
    sharding = jax.sharding.GSPMDSharding.get_replicated(jax.devices())
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, [self.ckpt_dir])

    serialize([arr], tspecs)

    # Run deserialization.
    (deserialized_arr,) = deserialize(
        shardings=[sharding],
        tensorstore_specs=tspecs,
        global_shapes=[shape],
        dtypes=[dtype],
    )

    out = deserialized_arr.astype(jnp.int8)  # doesn't crash
    self.assertEqual(out.dtype, jnp.int8)
    self.assertArraysEqual(out + out, out * 2)

  @parameterized.parameters((True,), (False,))
  def test_padding(self, strict: bool):
    data = np.arange(8)
    save_shape = data.shape
    global_mesh = create_global_mesh((2,), 'x')
    sharding = NamedSharding(global_mesh, P(None))
    array = jax.make_array_from_callback(
        save_shape, sharding, lambda idx: data[idx]
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([array], tspecs)

    restore_shape = (16,)
    if strict:
      with self.assertRaisesRegex(
          ValueError, 'is not compatible with the stored shape'
      ):
        deserialize([sharding], tspecs, [restore_shape], strict=strict)
    else:
      (restored,) = deserialize(
          [sharding], tspecs, [restore_shape], strict=strict
      )
      for shard in restored.addressable_shards:
        expected = np.arange(16)
        expected[8:] = 0
        self.assertArraysEqual(np.asarray(shard.data), expected)

  @parameterized.parameters((True,), (False,))
  def test_truncation(self, strict: bool):
    data = np.arange(16)
    save_shape = data.shape
    global_mesh = create_global_mesh((2,), 'x')
    sharding = NamedSharding(global_mesh, P(None))
    array = jax.make_array_from_callback(
        save_shape, sharding, lambda idx: data[idx]
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([array], tspecs)

    restore_shape = (8,)
    if strict:
      with self.assertRaisesRegex(
          ValueError, 'is not compatible with the stored shape'
      ):
        deserialize([sharding], tspecs, [restore_shape], strict=strict)
    else:
      (restored,) = deserialize(
          [sharding], tspecs, [restore_shape], strict=strict
      )
      for shard in restored.addressable_shards:
        self.assertArraysEqual(np.asarray(shard.data), np.arange(8))

  def test_odd_resharding(self):
    data = np.arange(12)
    global_shape = data.shape
    global_mesh = create_global_mesh((2,), 'x')
    sharding = NamedSharding(
        global_mesh,
        P(
            'x',
        ),
    )
    array = jax.make_array_from_callback(
        global_shape, sharding, lambda idx: data[idx]
    )
    ckpt_paths = [str(self.ckpt_dir)]
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, ckpt_paths)

    serialize([array], tspecs)

    global_mesh = create_global_mesh((3,), 'x')
    sharding = NamedSharding(
        global_mesh,
        P(
            'x',
        ),
    )
    (restored,) = deserialize([sharding], tspecs, [global_shape])
    for i, shard in enumerate(restored.addressable_shards):
      self.assertArraysEqual(np.asarray(shard.data), np.arange(4) + (i * 4))

  def test_load_with_layout(self):
    mesh = create_global_mesh((4, 2), ('x', 'y'))
    np_inp = np.arange(32).reshape(8, 4)
    s = NamedSharding(mesh, P('x', 'y'))
    arr = jax.device_put(np_inp, s)

    out_layout = (
        jax.jit(lambda x: x.T, out_shardings=Layout(DLL.AUTO))
        .lower(arr)
        .compile()
        .output_layouts
    )
    self.assertEqual(
        arr.layout.device_local_layout.major_to_minor,
        out_layout.device_local_layout.major_to_minor[::-1],
    )

    ckpt_dir = pathlib.Path(self.create_tempdir('ckpt').full_path)
    ckpt_path = pathlib.Path(self.create_tempdir(f'{ckpt_dir}/first').full_path)
    tspecs = jax.tree.map(serialization.get_tensorstore_spec, [ckpt_path])

    serialize(
        [arr],
        tspecs,
    )

    (out,) = deserialize([out_layout], tspecs)

    self.assertEqual(out.layout, out_layout)
    self.assertIsInstance(out, jax.Array)
    self.assertArraysEqual(out, np_inp)
    for s in out.addressable_shards:
      self.assertArraysEqual(s.data, np_inp[s.index])

  def test_incomplete_write(self):
    data = np.arange(8)
    chunk_len = 4
    global_mesh = create_global_mesh((8,), 'x')
    sharding = NamedSharding(global_mesh, P(None))
    tspec = ts_utils.ArrayWriteSpec(
        self.ckpt_dir.as_posix(),
        'a',
        global_shape=data.shape,
        write_shape=(chunk_len,),
        dtype=data.dtype,
        use_ocdbt=False,
    ).json
    t = ts.open(
        ts.Spec(tspec),
        create=True,
        open=True,
    ).result()
    t[:chunk_len].write(data[:chunk_len]).result()

    # Enable raising error for incomplete chunk.
    tspec['fill_missing_data_reads'] = False
    with self.assertRaisesRegex(
        Exception, 'Encountered error while reading array index'
    ):
      deserialize([sharding], [tspec])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses

from absl.testing import absltest
import jax
import numpy as np
from orbax.checkpoint._src.arrays import fragments as array_fragments

Fragment = array_fragments.Fragment
Fragments = array_fragments.Fragments


class FragmentTest(absltest.TestCase):

  def test_construction(self):
    with self.subTest('rejects_index_as_ndarray'):
      index = np.s_[1:2:1, 3:4:1]
      with self.assertRaisesRegex(TypeError, 'must be a tuple of slices'):
        _ = Fragment(index=array_fragments._ndarray_from_index(index))  # pytype: disable=wrong-arg-types

    with self.subTest('with_np_index'):
      f = Fragment(index=np.s_[1:2:1, 3:4:1])
      self.assertEqual(f, Fragment(np_index=f.np_index))

    with self.subTest('with_dataclasses_replace_value'):
      f_a = Fragment(index=np.s_[0:2:1, 0:3:1])
      value = np.ones((2, 3))
      f_c = dataclasses.replace(f_a, value=value)
      self.assertEqual(f_c, Fragment(index=f_a.index, value=value))

    with self.subTest('with_dataclasses_replace_np_index'):
      value = np.ones((2, 3))
      old_np_index = array_fragments._ndarray_from_index(np.s_[0:2:1, 0:3:1])
      new_np_index = array_fragments._ndarray_from_index(np.s_[2:4:1, 3:6:1])
      f_old = Fragment(np_index=old_np_index, value=value)
      f_new = dataclasses.replace(f_old, np_index=new_np_index)
      self.assertEqual(f_new, Fragment(np_index=new_np_index, value=value))

    with self.subTest('only_one_of_index_or_np_index_is_allowed'):
      idx = np.s_[1:2:1, 3:4:1]
      with self.assertRaisesRegex(ValueError, 'both index and np_index'):
        _ = Fragment(
            index=idx, np_index=array_fragments._ndarray_from_index(idx)
        )

    with self.subTest('one_of_index_or_np_index_must_be_specified'):
      with self.assertRaisesRegex(ValueError, 'either index or np_index'):
        _ = Fragment()

  def test_equality(self):
    self.assertEqual(
        Fragment(index=np.s_[1:2:1, 3:4:1]),
        Fragment(index=np.s_[1:2:1, 3:4:1]),
    )
    self.assertNotEqual(
        Fragment(index=np.s_[1:2:1, 3:4:1]),
        Fragment(index=np.s_[1:2:1, 4:5:1]),
    )
    self.assertNotEqual(
        Fragment(index=np.s_[1:2:1, 3:4:1], value=np.ones([4, 4]) * 2.0),
        Fragment(index=np.s_[1:2:1, 3:4:1], value=np.ones([4, 4]) * 3.0),
    )

  def test_nd_properties(self):
    f = Fragment(index=np.s_[1:8:2, 3:9:4])
    np.testing.assert_array_equal([1, 3], f.start)
    np.testing.assert_array_equal([8, 9], f.stop)
    np.testing.assert_array_equal([2, 4], f.step)

  def test_rank_zero(self):
    f = Fragment(index=(), value=np.empty([]))
    np.testing.assert_array_equal([], f.start)
    np.testing.assert_array_equal([], f.stop)
    np.testing.assert_array_equal([], f.step)

  def test_repr(self):
    self.assertEqual(
        repr(Fragment(index=np.s_[1:2:1, 3:4:1])),
        'Fragment(index=np.s_[1:2:1, 3:4:1])',
    )
    self.assertEqual(
        repr(Fragment(index=np.s_[1:2:1, 3:4:1], value=np.ones([4, 4]))),
        'Fragment(index=np.s_[1:2:1, 3:4:1], value=...)',
    )

  def test_nbytes_of_concrete_fragment_is_nbytes_of_its_value(self):
    fragment_value = np.full([4, 4], 2.0)
    self.assertEqual(
        fragment_value.nbytes,
        Fragment(index=np.s_[1:5:1, 3:7:1], value=fragment_value).nbytes,
    )

  def test_nbytes_of_abstract_fragment_raises_value_error(self):
    with self.assertRaisesRegex(ValueError, 'nbytes of abstract'):
      _ = Fragment(index=np.s_[1:5:1, 3:7:1]).nbytes

  def test_nbytes_astype_of_concrete_fragment_uses_given_dtype(self):
    fragment_value = np.full([4, 4], 2.0)
    self.assertEqual(
        fragment_value.astype(np.int8).nbytes,
        Fragment(index=np.s_[1:5:1, 3:7:1], value=fragment_value).nbytes_astype(
            np.dtype(np.int8)
        ),
    )

  def test_nbytes_astype_of_abstract_fragment_uses_given_dtype(self):
    self.assertEqual(
        4 * 4 * 2,
        Fragment(index=np.s_[1:5:1, 3:7:1]).nbytes_astype(
            np.dtype(jax.numpy.bfloat16)
        ),
    )


class FragmentsTest(absltest.TestCase):

  def test_can_be_constructed_with_fragment_list(self):
    Fragments(
        shape=(4, 4),
        dtype=np.dtype(np.float32),
        fragments=[Fragment(index=np.s_[1:2:1, 3:4:1])],
    )

  def test_cannot_be_constructed_with_non_fragment_list(self):
    with self.assertRaises(TypeError):
      Fragments(
          shape=(4, 4),
          dtype=np.dtype(np.float32),
          fragments=[np.s_[1:2:1, 3:4:1]],
      )

  def test_non_degenerate_fragments(self):
    fragments = Fragments(
        shape=(5, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[1:2:1, 3:4:1]),
            Fragment(index=np.s_[3:5:1, 0:2:1]),
        ],
    )
    self.assertFalse(fragments.is_degenerate())

  def test_non_degenerate_fragments_due_to_one_non_empty_fragment(self):
    fragments = Fragments(
        shape=(5, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[1:2:1, 3:4:1]),
            Fragment(index=np.s_[3:3:1, 0:2:1]),
        ],
    )
    self.assertFalse(fragments.is_degenerate())

  def test_non_degenerate_fragments_due_to_all_degenerate_fragment(self):
    fragments = Fragments(
        shape=(5, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[2:2:1, 3:4:1]),
            Fragment(index=np.s_[3:5:1, 0:0:1]),
        ],
    )
    self.assertTrue(fragments.is_degenerate())

  def test_nbytes_of_fragments_matches_nbytes_of_same_shaped_array(self):
    fragments = Fragments(
        shape=(5, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[1:2:1, 3:4:1]),
            Fragment(index=np.s_[3:5:1, 0:2:1]),
        ],
    )
    self.assertEqual(np.ones((5, 5), np.float32).nbytes, fragments.nbytes)

  def test_addressable_nbytes_of_abstract_fragments_is_sum_of_individual_fragment_bytes(
      self,
  ):
    fragments = Fragments(
        shape=(5, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[1:2:1, 3:4:1]),
            Fragment(index=np.s_[3:5:1, 0:2:1]),
        ],
    )
    self.assertEqual(((1 * 1) + (2 * 2)) * 4, fragments.addressable_nbytes)


class FragmentsToArrayTest(absltest.TestCase):

  def test_full_fragments_can_be_converted_to_numpy_array(self):
    fragments = Fragments(
        shape=(4, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[0:4:1, 0:2:1], value=np.full([4, 2], 88)),
            Fragment(index=np.s_[0:4:1, 2:5:1], value=np.full([4, 3], 99)),
        ],
    )
    a = np.asarray(fragments)
    np.testing.assert_array_equal(
        np.concatenate([np.full([4, 2], 88), np.full([4, 3], 99)], axis=1), a
    )

  def test_non_full_fragments_raises_exception(self):
    fragments = Fragments(
        shape=(4, 5),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[0:4:1, 0:2:1], value=np.full([4, 2], 88))
        ],
    )
    with self.assertRaisesRegex(
        ValueError,
        r'Attempt to convert non-full Fragments to array',
    ):
      np.asarray(fragments)


class IsFullTest(absltest.TestCase):

  def test_filled_by_one_spanning_fragment(self):
    self.assertTrue(
        array_fragments._is_full(
            Fragments(
                shape=(4, 5),
                dtype=np.dtype(np.float32),
                fragments=[
                    Fragment(index=np.s_[0:4:1, 0:5:1]),
                ],
            )
        )
    )

  def test_filled_by_two_non_spanning_fragments(self):
    self.assertTrue(
        array_fragments._is_full(
            Fragments(
                shape=(4, 5),
                dtype=np.dtype(np.float32),
                fragments=[
                    Fragment(index=np.s_[0:4:1, 0:2:1]),
                    Fragment(index=np.s_[0:4:1, 2:5:1]),
                ],
            )
        )
    )

  def test_not_filled_by_two_non_spanning_fragments(self):
    self.assertFalse(
        array_fragments._is_full(
            Fragments(
                shape=(4, 5),
                dtype=np.dtype(np.float32),
                fragments=[
                    Fragment(index=np.s_[0:4:1, 0:2:1]),
                    Fragment(index=np.s_[2:4:1, 2:5:1]),
                ],
            )
        )
    )

  def test_not_filled_by_no_fragments(self):
    self.assertFalse(
        array_fragments._is_full(
            Fragments(shape=(4, 5), dtype=np.dtype(np.float32), fragments=[])
        )
    )

  def test_rank_0_fragments_not_filled_by_no_fragments(self):
    self.assertFalse(
        array_fragments._is_full(
            Fragments(shape=(), dtype=np.dtype(np.float32), fragments=[])
        )
    )

  def test_dim_0_fragments_filled_by_no_fragments(self):
    self.assertTrue(
        array_fragments._is_full(
            Fragments(shape=(0,), dtype=np.dtype(np.float32), fragments=[])
        )
    )


class AddressableShardsTest(absltest.TestCase):

  def test_unsharded_array_is_fully_replicated(self):
    self.assertEqual(
        array_fragments.addressable_shards(
            jax.ShapeDtypeStruct((4, 5), np.dtype(np.float32))
        ),
        [np.s_[0:4:1, 0:5:1]],
    )


class AbstractFragmentsTest(absltest.TestCase):

  def test_returns_fragments_instance_itself(self):
    fragments = Fragments(
        shape=(2, 3), dtype=np.dtype(np.float32), fragments=[]
    )
    self.assertIs(fragments, array_fragments.abstract_fragments(fragments))

  def test_converts_fully_replicated_shape_dtype_struct(self):
    self.assertEqual(
        Fragments(
            shape=(4, 5),
            dtype=np.dtype(np.float32),
            fragments=[Fragment(index=np.s_[0:4:1, 0:5:1])],
        ),
        array_fragments.abstract_fragments(
            jax.ShapeDtypeStruct((4, 5), np.dtype(np.float32))
        ),
    )


class StackFragmentsTest(absltest.TestCase):

  def test_stacks_fragments_of_same_shape(self):
    fragments = Fragments(
        shape=(24, 35),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[12:16:1, 10:15:1], value=np.full((4, 5), 1.0)),
            Fragment(index=np.s_[16:20:1, 10:15:1], value=np.full((4, 5), 2.0)),
            Fragment(index=np.s_[12:16:1, 10:15:1], value=np.full((4, 5), 3.0)),
            Fragment(index=np.s_[16:20:1, 10:15:1], value=np.full((4, 5), 4.0)),
        ],
    )

    expected_array = np.stack([
        np.full((4, 5), 1.0),
        np.full((4, 5), 2.0),
        np.full((4, 5), 3.0),
        np.full((4, 5), 4.0),
    ])

    actual_array = array_fragments.stack_fragments(fragments)
    np.testing.assert_array_equal(expected_array, actual_array)

  def test_stacks_single_fragment_without_copy(self):
    value = np.arange(20).reshape((4, 5))
    fragments = Fragments(
        shape=(4, 5),
        dtype=np.dtype(np.float32),
        fragments=[Fragment(index=np.s_[0:4:1, 0:5:1], value=value)],
    )

    actual_array = array_fragments.stack_fragments(fragments)
    assert actual_array is not None
    self.assertEqual(actual_array.shape, (1, 4, 5))
    # Result is a view of the original `np.arange(20)` array.
    self.assertIs(actual_array.base, value.base)
    np.testing.assert_array_equal(
        actual_array, np.arange(20).reshape((1, 4, 5))
    )

  def test_returns_none_for_none(self):
    self.assertIsNone(array_fragments.stack_fragments(None))

  def assert_stacking_is_rejected(
      self,
      invalid_fragments: Fragments,
      expected_error_message: str,
  ) -> None:
    with self.assertRaisesRegex(ValueError, expected_error_message):
      array_fragments.validate_fragments_can_be_stacked(invalid_fragments)
    with self.assertRaisesRegex(ValueError, expected_error_message):
      array_fragments.stack_fragments(invalid_fragments)

  def test_rejects_empty_fragments_list(self):
    empty_fragments = Fragments(
        shape=(24, 35), dtype=np.dtype(np.float32), fragments=[]
    )
    self.assert_stacking_is_rejected(empty_fragments, 'No fragments to stack')

  def test_rejects_fragments_of_different_shapes(self):
    fragments = Fragments(
        shape=(24, 35),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[12:16:1, 10:15:1], value=np.full((4, 5), 1.0)),
            Fragment(
                index=np.s_[16:20:1, 10:20:1], value=np.full((4, 10), 2.0)
            ),
        ],
    )
    self.assert_stacking_is_rejected(fragments, 'Differently-shaped fragments')

  def test_rejects_abstract_fragments(self):
    fragments = Fragments(
        shape=(24, 35),
        dtype=np.dtype(np.float32),
        fragments=[
            Fragment(index=np.s_[12:16:1, 10:15:1], value=np.full((4, 5), 1.0)),
            Fragment(index=np.s_[16:20:1, 15:20:1], value=None),
        ],
    )

    self.assert_stacking_is_rejected(fragments, 'Not all fragments have values')


if __name__ == '__main__':
  absltest.main()
